/*
 * Copyright (c) 2009 Serg79
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* MFC   */
#define WIN32_LEAN_AND_MEAN

/*   UNICODE */
#define UNICODE        /* Windows */

#ifdef UNICODE
#define _UNICODE       /* RTL */
#endif

#include <windows.h>
#include <Tchar.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#define DEF_BUFF_SIZE        1024

#if DEF_BUFF_SIZE < MAX_PATH
#undef DEF_BUFF_SIZE
#define DEF_BUFF_SIZE    MAX_PATH
#endif

#define DEF_LPBUFF_SIZE        20

#ifdef _DEBUG
#define LogDebug(a)   do { MessageBox(NULL, (a), _T("LogDebug"), MB_OK); } while(0)
#else
#define LogDebug(a)
#endif

static TCHAR *szProgName;
static TCHAR szProgDirectory[DEF_BUFF_SIZE];

static LPCTSTR szLogFileName = _T("duplicate.log");
static LPCTSTR szListFileName = _T("duplicate.lst");
static LPCTSTR szLockFileName = _T("duplicate.lock");

static BOOL flagLog = FALSE;
static BOOL flagExplorer = TRUE;

static void *xmalloc(size_t size)
{
	void *p;
	
	if ((p = malloc(size)) == NULL) {
		LogDebug(_T("Failed allocate memory, function 'xmalloc'."));
		exit(1);
	}

	return p;
}

static void *xcalloc(size_t size)
{
	void *p;

	if ((p = calloc(1, size)) == NULL) {
		LogDebug(_T("Failed allocate memory, function 'xcalloc'."));
		exit(1);
	}

	return p;
}

static void *xrealloc(void *ptr, size_t size)
{
	void *p;

	if ((p = realloc(ptr, size)) == NULL) {
		LogDebug(_T("Failed allocate memory, function 'xrealloc'."));
		exit(1);
	}

	return p;
}

static void xfree(void *p)
{
	free(p);
}

#ifdef UNICODE

/*    UNICODE  ANSI */
static LPSTR GetAnsiString(LPCWSTR lpString)
{
	static LPSTR lpBuff = NULL;
	static int lenBuff = 0;
	int lenString;

	if (!lpBuff) {
		lpBuff = xmalloc(sizeof(lpBuff[0]) * DEF_BUFF_SIZE);
		lenBuff = DEF_BUFF_SIZE;
	}

	lenString = WideCharToMultiByte(CP_ACP, 0, lpString, -1, NULL, 0, NULL, NULL);
	while (lenString >= lenBuff) {
		lenBuff += DEF_BUFF_SIZE;
		lpBuff = xrealloc(lpBuff, sizeof(lpBuff[0]) * lenBuff);
	}

	WideCharToMultiByte(CP_ACP, 0, lpString, -1, lpBuff, lenBuff, NULL, NULL);
	return lpBuff;
}

/*    ANSI  UNICODE */
static LPWSTR GetWideString(LPCSTR lpString)
{
	static LPWSTR lpBuff = NULL;
	static int lenBuff = 0;
	int lenString;

	if (!lpBuff) {
		lpBuff = xmalloc(sizeof(lpBuff[0]) * DEF_BUFF_SIZE);
		lenBuff = DEF_BUFF_SIZE;
	}

	lenString = MultiByteToWideChar(CP_ACP, 0, lpString, -1, NULL, 0);
	while (lenString >= lenBuff) {
		lenBuff += DEF_BUFF_SIZE;
		lpBuff = xrealloc(lpBuff, sizeof(lpBuff[0]) * lenBuff);
	}

	MultiByteToWideChar(CP_ACP, 0, lpString, -1, lpBuff, lenBuff);
	return lpBuff;
}

#endif  /* UNICODE */

/*  LocalWriteFile     ANSI */
static void LocalWriteFile(FILE *fFile, LPCSTR lpString)
{
	if (fputs(lpString, fFile) == EOF) {
		LogDebug(_T("Runtime error 'WriteFile' in 'LocalWriteFile'."));
		exit(1);
	}
}

/*  LogMessage   log-    ANSI */
static void LogMessage(LPCTSTR lpMessage, ...)
{
	static FILE *fLogFile = NULL;
	static LPTSTR lpBuff = NULL;
	static int lenBuff = 0;
	LPSTR lpTime;
	LPCSTR lpString;
	va_list args;
	time_t tm;

	if (!flagLog)
		return;

	if (!fLogFile) {
		fLogFile = _tfopen(szLogFileName, _T("a+"));
		if (fLogFile == NULL) {
			LogDebug(_T("Runtime error 'CreateFile' in 'LogMessage'."));
			exit(1);
		}
		lpBuff = xmalloc(sizeof(lpBuff[0]) * DEF_BUFF_SIZE);
		lenBuff = DEF_BUFF_SIZE;
	}

	tm = time(NULL);
	lpTime = ctime(&tm);
	lpTime[lstrlenA(lpTime) - 1] = '\0';
	LocalWriteFile(fLogFile, lpTime);

	LocalWriteFile(fLogFile, "# ");

	va_start(args, lpMessage);
	while (_vsntprintf(lpBuff, lenBuff, lpMessage, args) == -1) {
		lenBuff += DEF_BUFF_SIZE;
		lpBuff = xrealloc(lpBuff, sizeof(lpBuff[0]) * lenBuff);
	}
	va_end(args);
#ifdef UNICODE
	lpString = GetAnsiString(lpBuff);
#else
	lpString = lpBuff;
#endif
	LocalWriteFile(fLogFile, lpString);

	LocalWriteFile(fLogFile, "\n");
}

void RunExplorer(void)
{
	TCHAR lpBuff[56];
	STARTUPINFO si = { sizeof(si) };
	PROCESS_INFORMATION pi;
	BOOL fSuccess;

	if (!flagExplorer)
		return;

	lpBuff[0] = _T('\0');
	lstrcat(lpBuff, _T("explorer "));
	lstrcpyn(&lpBuff[lstrlen(lpBuff)], szProgName, 3);

	fSuccess = CreateProcess(NULL, lpBuff, NULL, NULL,
		FALSE, 0, NULL, NULL, &si, &pi);
	if (fSuccess) {
		CloseHandle(pi.hThread);
		CloseHandle(pi.hProcess);
	} else {
		LogDebug(_T("Runtime error 'CreateProcess' in 'RunExplorer'."));
		exit(1);
	}
}

static void VersionCheckNT(void)
{
	OSVERSIONINFO os = { sizeof(os) };

	if (!GetVersionEx(&os)) {
		LogMessage(_T("Runtime error (%d) 'GetVersionEx' in 'VersionCheckNT'."), GetLastError());
		exit(1);
	}

	if (os.dwPlatformId != VER_PLATFORM_WIN32_NT) {
		LogMessage(_T("Need NT version of Windows."));
		exit(1);
	}
}

static LPTSTR GetCopyFileName(FILE *fp)
{
	static CHAR lpBuff[DEF_BUFF_SIZE];
	BOOL fEof = FALSE;
	int count;

	for (count = 0; count < (DEF_BUFF_SIZE - 1); count++) {
		int ch = fgetc(fp);
		if (ch == EOF) {
			fEof = TRUE;
			break;
		} else if (ch == '\n') {
			break;
		}
		lpBuff[count] = ch;
	}
	if (fEof && !count)
		return NULL;

	lpBuff[count] = '\0';

#ifdef UNICODE
	return GetWideString(lpBuff);
#else
	return lpBuff;
#endif
}

static BOOL CreatePath(LPCTSTR lpLocalPath)
{
	TCHAR lpPath[DEF_BUFF_SIZE];
	DWORD dwError;
	int i;

	for (i = 0; lpLocalPath[i]; i++) {
		if (lpLocalPath[i] == _T('\\')) {
			lstrcpyn(lpPath, lpLocalPath, i + 1);
			if (!CreateDirectory(lpPath, NULL)) {
				dwError = GetLastError();
				if (dwError != ERROR_ALREADY_EXISTS) {
					LogMessage(_T("Runtime error (%d) 'CreateDirectory' in 'CreatePath': %s"),
						dwError, lpLocalPath);
					return FALSE;
				}
			}
		}
	}

	return TRUE;
}

static void LocalCopyFile(LPCTSTR lpFullName, LPWIN32_FIND_DATA lpFileData, int width)
{
#define MODE_NEW       0
#define MODE_UPDATE    1
#define MODE_ACTUAL    2
#define MODE_ERROR     3
#define MODE_FORGET    4
	static const TCHAR *lpMode[] = {
		_T("new"),
		_T("update"),
		_T("actual"),
		_T("error"),
		_T("forget")
	};
	struct {
		WORD FatDate;
		WORD FatTime;
	} dtCreation, dtWrite, dtTempCreation, dtTempWrite;
	FILETIME ftCreation, ftWrite;
	HANDLE hFile;
	TCHAR lpDuplicate[DEF_BUFF_SIZE];
	BOOL fConvertTimeError = FALSE;
	int mode = MODE_FORGET;

	lstrcpy(lpDuplicate, lpFullName);
	if (lpDuplicate[1] != _T(':')) {
		LogMessage(_T("The drive is not correctly: '%s'"), lpFullName);
		return;
	} else {
		lpDuplicate[1] = _T('_');
	}

	hFile = CreateFile(lpDuplicate, 0, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		DWORD dwError = GetLastError();

		if (dwError == ERROR_FILE_NOT_FOUND) {
			mode = MODE_NEW;
		} else if (dwError == ERROR_PATH_NOT_FOUND) {
			mode = CreatePath(lpDuplicate) ? MODE_NEW : MODE_ERROR;
		} else {
			LogMessage(_T("Runtime error (%d) 'CreateFile1' in 'LocalCopyFile': %s"),
				GetLastError(), lpFullName);
			mode = MODE_ERROR;
		}
	} else {
		if (GetFileTime(hFile, &ftCreation, NULL, &ftWrite)) {
			/*
			 *   FAT        NTFS,
			 *      FAT .
			 */
			if (!FileTimeToDosDateTime(&lpFileData->ftCreationTime,
					&dtCreation.FatDate, &dtCreation.FatTime)) {
				LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime1' in 'LocalCopyFile': %s"),
					GetLastError(), lpFullName);
				fConvertTimeError = TRUE;
			}
			if (!FileTimeToDosDateTime(&lpFileData->ftLastWriteTime,
					&dtWrite.FatDate, &dtWrite.FatTime)) {
				LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime2' in 'LocalCopyFile': %s"),
					GetLastError(), lpFullName);
				fConvertTimeError = TRUE;
			}
			if (!fConvertTimeError) {
				if (!FileTimeToDosDateTime(&ftCreation,
						&dtTempCreation.FatDate, &dtTempCreation.FatTime)) {
					LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime3' in 'LocalCopyFile':%s"),
						GetLastError(), lpFullName);
					fConvertTimeError = TRUE;
				}
				if (!FileTimeToDosDateTime(&ftWrite,
						&dtTempWrite.FatDate, &dtTempWrite.FatTime)) {
					LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime4' in 'LocalCopyFile': %s"),
						GetLastError(), lpFullName);
					fConvertTimeError = TRUE;
				}
			}

			if (!fConvertTimeError) {
				if (((dtCreation.FatDate == dtTempCreation.FatDate)
						&& (dtCreation.FatTime == dtTempCreation.FatTime))
						&& ((dtWrite.FatDate == dtTempWrite.FatDate)
						&& (dtWrite.FatTime == dtTempWrite.FatTime))) {
					mode = MODE_ACTUAL;
				} else {
					mode = MODE_UPDATE;
				}
			} else {
				mode = MODE_UPDATE;
			}
		} else {
			LogMessage(_T("Runtime error (%d) 'GetFileTime' in 'LocalCopyFile': %s"),
				GetLastError(), lpFullName);
			mode = MODE_ERROR;
		}

		CloseHandle(hFile);
	}

	if (mode == MODE_NEW || mode == MODE_UPDATE) {
		if (CopyFile(lpFullName, lpDuplicate, FALSE)) {
			hFile = CreateFile(lpDuplicate, GENERIC_WRITE, 0, NULL,
				OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if (hFile != INVALID_HANDLE_VALUE) {
				if (mode == MODE_NEW) {
					if (!FileTimeToDosDateTime(&lpFileData->ftCreationTime,
							&dtCreation.FatDate, &dtCreation.FatTime)) {
						LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime5' in 'LocalCopyFile': %s"),
							GetLastError(), lpFullName);
						fConvertTimeError = TRUE;
					}
					if (!FileTimeToDosDateTime(&lpFileData->ftLastWriteTime,
							&dtWrite.FatDate, &dtWrite.FatTime)) {
						LogMessage(_T("Runtime error (%d) 'FileTimeToDosDateTime6' in 'LocalCopyFile': %s"),
							GetLastError(), lpFullName);
						fConvertTimeError = TRUE;
					}
				}

				if (!fConvertTimeError) {
					if (!DosDateTimeToFileTime(dtCreation.FatDate, dtCreation.FatTime, &ftCreation)) {
						LogMessage(_T("Runtime error (%d) 'DosDateTimeToFileTime1' in 'LocalCopyFile': %s"),
							GetLastError(), lpFullName);
						fConvertTimeError = TRUE;
					}
					if (!DosDateTimeToFileTime(dtWrite.FatDate, dtWrite.FatTime, &ftWrite)) {
						LogMessage(_T("Runtime error (%d) 'DosDateTimeToFileTime2' in 'LocalCopyFile': %s"),
							GetLastError(), lpFullName);
						fConvertTimeError = TRUE;
					}

					if (!fConvertTimeError) {
						if (!SetFileTime(hFile, &ftCreation, NULL, &ftWrite)) {
							LogMessage(_T("Runtime error (%d) 'SetFileTime' in 'LocalCopyFile': %s"),
								GetLastError(), lpFullName);
							mode = MODE_ERROR;
						}
					} else {
						mode = MODE_ERROR;
					}
				}
				CloseHandle(hFile);
			} else {
				LogMessage(_T("Runtime error (%d) 'CreateFile2' in 'LocalCopyFile': %s"),
					GetLastError(), lpFullName);
				mode = MODE_ERROR;
			}
		} else {
			LogMessage(_T("Runtime error (%d) 'CopyFile' in 'LocalCopyFile': %s"),
				GetLastError(), lpFullName);
			mode = MODE_ERROR;
		}
	}

	LogMessage(_T("%*cf][%s]%s"), width, _T('['), lpMode[mode], lpFullName);
#undef MODE_NEW
#undef MODE_UPDATE
#undef MODE_ACTUAL
#undef MODE_ERROR
#undef MODE_FORGET
}

#define DOPATH_COPY             0      /*  */
#define DOPATH_COPY_RECURSIVE   1      /*   */
#define DOPATH_LIST             2      /*    */
#define DOPATH_LIST_RECURSIVE   3      /*     */

static void DoPath(LPCTSTR lpPattern, int mode, int width)
{
	TCHAR lpFullName[DEF_BUFF_SIZE];
	WIN32_FIND_DATA FileData;
	HANDLE hSearch;
	int offset;
	BOOL fFinished = FALSE;

	hSearch = FindFirstFile(lpPattern, &FileData);
	if (hSearch == INVALID_HANDLE_VALUE) {
		LogMessage(_T("Not found: '%s'"), lpPattern);
		return;
	}

	offset = _tcsrchr(lpPattern, _T('\\')) - lpPattern + 1;
	lstrcpyn(lpFullName, lpPattern, offset + 1);

	while (!fFinished) {
		TCHAR *lpName = FileData.cFileName;
		lstrcpyn(lpFullName + offset, lpName, DEF_BUFF_SIZE - offset);

		if (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			if (lstrcmp(lpName, _T(".")) && lstrcmp(lpName, _T(".."))) {
				static const TCHAR lpTail[] = _T("\\*");
				int lenFullName;

				if (mode == DOPATH_COPY_RECURSIVE || mode == DOPATH_LIST_RECURSIVE) {
					LogMessage(_T("%*cd]%s\\"), width + 1, _T('['), lpFullName);

					lenFullName = lstrlen(lpFullName);
					lstrcpyn(lpFullName + lenFullName, lpTail,
						DEF_BUFF_SIZE - lenFullName - lstrlen(lpTail));
					DoPath(lpFullName, mode, width + 1);
				} else {
					LogMessage(_T("%*cd]%s\\"), width, _T('['), lpFullName);
				}
			}
		} else {
			if (mode == DOPATH_LIST || mode == DOPATH_LIST_RECURSIVE)
				LogMessage(_T("%*cf]%s"), width, _T('['), lpFullName);
			else
				LocalCopyFile(lpFullName, &FileData, width);
		}

		if (!FindNextFile(hSearch, &FileData)) {
			if (GetLastError() != ERROR_NO_MORE_FILES)
				LogMessage(_T("Runtime error (%d) 'FindNextFile' in 'DoPath'."), GetLastError());
			fFinished = TRUE;
		}
	}

	FindClose(hSearch);
}

#ifdef _CONSOLE
int _tmain(void)
#elif _WINDOWS
int WINAPI _tWinMain(HINSTANCE hInst, HINSTANCE hPrevInst, PTSTR lpCmdLine, int nCmdShow)
#else
#error Unspecified type of program...
#endif
{
	TCHAR szMutexName[DEF_BUFF_SIZE];
	TCHAR **argv, **lpListFile;
	LPTSTR lpBuff;
	FILE *fpListFile;
	HANDLE hMutex;
	int argc, lenListFile, i;

	/*     */
	argc = __argc;
#ifdef UNICODE
	argv = __wargv;
#else
	argv = __argv;
#endif
	szProgName = argv[0];

	for (i = 1; i < argc; i++) {
		if (!lstrcmp(argv[i], _T("--enable-log"))) {
			flagLog = TRUE;
		}
		else if (!lstrcmp(argv[i], _T("--disable-explorer"))) {
			flagExplorer = FALSE;
		}
	}

	/*  explorer */
	RunExplorer();

	/*       */
	for (i = 0; szProgName[i]; i++) {
		TCHAR ch = szProgName[i];
		szMutexName[i] = (ch == _T(':') || ch == _T('\\')) ? _T('_') : ch;
	}
	szMutexName[i] = _T('\0');
	hMutex = CreateMutex(NULL, FALSE, szMutexName);
	if (GetLastError() == ERROR_ALREADY_EXISTS) {
		return 0;
	}

	/*      */
	i = _tcsrchr(szProgName, _T('\\')) - szProgName + 1;
	lstrcpyn(szProgDirectory, szProgName, i + 1);
	if (!SetCurrentDirectory(szProgDirectory)) {
		LogDebug(_T("Error when changing the current directory."));
		exit(1);
	}

	LogMessage(_T("######################################"));
	LogMessage(_T("Running '%s'"), szProgName);

	/*   OS */
	VersionCheckNT();

	/* C   . */
	fpListFile = _tfopen(szListFileName, _T("r"));
	if (fpListFile == NULL) {
		LogMessage(_T("Can not find the list '%s' of copied files."), szListFileName);
		exit(1);
	}

	lpListFile = xcalloc(sizeof(TCHAR *) * DEF_LPBUFF_SIZE);
	lenListFile = DEF_LPBUFF_SIZE;

	for (i = 0; (lpBuff = GetCopyFileName(fpListFile)) != NULL;) {
		int lenBuff;

		if (!lpBuff[0] || (lpBuff[0] == _T('#')))
			continue;

		lenBuff = lstrlen(lpBuff) + 1;
		lpListFile[i] = xmalloc(sizeof(TCHAR) * lenBuff);
		lstrcpyn(lpListFile[i], lpBuff, lenBuff);
		if (lpListFile[i][lenBuff - 2] == _T('\\'))
			lpListFile[i][lenBuff - 2] = _T('\0');

		LogMessage(_T("Add pattern: '%s'"), lpBuff);

		if (++i >= lenListFile) {
			lpListFile = xrealloc(lpListFile, sizeof(TCHAR *) * (lenListFile + DEF_LPBUFF_SIZE));
			ZeroMemory(lpListFile + lenListFile, sizeof(TCHAR *) * DEF_LPBUFF_SIZE);
			lenListFile += DEF_LPBUFF_SIZE;
		}
	}

	fclose(fpListFile);
	LogMessage(_T("======================================"));

	/*   */
	for (i = 0; lpListFile[i]; i++) {
		TCHAR *lpList = lpListFile[i];
		int mode;

		LogMessage(_T("pattern: %s"), lpList);

		if (lpList[0] == _T('d') && lpList[1] == _T('=')) {
			TCHAR lpBuff[DEF_BUFF_SIZE];

			if (GetLogicalDriveStrings(DEF_BUFF_SIZE, lpBuff)) {
				int i;

				for (i = 0; ; i++) {
					if (!lpBuff[i]) {
						if (lpBuff[i + 1])
							lpBuff[i] = _T(' ');
						else
							break;
					}
				}

				LogMessage(_T("drives: %s"), lpBuff);
			} else {
				LogMessage(_T("Runtime error (%d) 'GetLogicalDriveStrings' in 'main'."),
					GetLastError());
			}

			LogMessage(_T("======================================"));
			continue;
		} else if (lpList[0] == _T('c') && lpList[1] == _T('=')) {
			mode = DOPATH_COPY;
			lpList += 2;
		} else if (lpList[0] == _T('C') && lpList[1] == _T('=')) {
			mode = DOPATH_COPY_RECURSIVE;
			lpList += 2;
		} else if (lpList[0] == _T('l') && lpList[1] == _T('=')) {
			mode = DOPATH_LIST;
			lpList += 2;
		} else if (lpList[0] == _T('L') && lpList[1] == _T('=')) {
			mode = DOPATH_LIST_RECURSIVE;
			lpList += 2;
		} else {
			mode = DOPATH_COPY;
		}

		DoPath(lpList, mode, 1);

		LogMessage(_T("======================================"));
	}

	return 0;
}
